#include <stdio.h>
#include <stdlib.h>

#include "main.h"
#define BUFFER_SIZE 1024

#define USAGE "Usage: %s <pspbtcnf file> -build | -extract -b | -e\nor %s -v\n-e | -extract: extract pspbtcnf file to pspbcnf.txt\n-b | -build: build pspbtcnf file from pspbtcnf.txt\n-v | -version: version info\n"
#define VERSION_NUMBER "0.2"
#define BUILD_DATE "3 March 2008"

#define BTCNF "pspbtcnf.txt"
#define BTCNF_VSH "pspbtcnf_vsh.txt"
#define BTCNF_GAME "pspbtcnf_game.txt"
#define BTCNF_UPDATER "pspbtcnf_updater.txt"
#define BTCNF_POPS "pspbtcnf_pops.txt"
#define BTCNF_APP "pspbtcnf_app.txt"

long fcopy(char *source, char *dest);
int main(int argc, char **argv);
int split_btcnf(char *oldfile);
int build_btcnf(char *oldfile);
int check_sig(char *infile);

long fcopy(char *source, char *dest)
{
        FILE *d, *s;
        char *buffer;
        size_t incount;
        long totcount = 0L;

        s = fopen(source, "rb");
        if(s == NULL)
                return -1L;

        d = fopen(dest, "wb");
        if(d == NULL)
        {
                fclose(s);
                return -1L;
        }

        buffer = malloc(BUFFER_SIZE);
        if(buffer == NULL)
        {
                fclose(s);
                fclose(d);
                return -1L;
        }

        incount = fread(buffer, sizeof(char), BUFFER_SIZE, s);

        while(!feof(s))
        {
                totcount += (long)incount;
                fwrite(buffer, sizeof(char), incount, d);
                incount = fread(buffer, sizeof(char), BUFFER_SIZE, s);
        }

        totcount += (long)incount;
        fwrite(buffer, sizeof(char), incount, d);

        free(buffer);
        fclose(s);
        fclose(d);

        return totcount;
}

int main(int argc, char **argv)
{
	if(argc < 2) return !(printf(USAGE, argv[0], argv[0])+(strncmp(argv[0], "/cygdrive", 9) ? 0 : getc(stdin)))-1;

	if(argc == 2)
		if(strcmp(argv[1], "-version") == 0 || strcmp(argv[1], "-v") == 0)
			return !printf("NewBTCNFedit version " VERSION_NUMBER ", " BUILD_DATE "\n");
		else return !printf(USAGE, argv[0], argv[0])-1;
	
	if(strcmp(argv[2], "-build") && strcmp(argv[2], "-extract") && strcmp(argv[1], "-version") && strcmp(argv[2], "-b") && strcmp(argv[2], "-e") && strcmp(argv[1], "-v"))
		return !printf(USAGE, argv[0], argv[0])-1;

	if(strcmp(argv[2], "-build") == 0 || strcmp(argv[2], "-b") == 0)
		return build_btcnf(argv[1]);
	else if(strcmp(argv[2], "-extract") == 0 || strcmp(argv[2], "-e") == 0)
			return split_btcnf(argv[1]);
	else return 0;
}

int check_sig(char *infile)
{
	FILE *in;
	u32 sig;
	in = fopen(infile, "r");
	if(!in)
	{
		return !printf("Input file %s not found!\n", infile)-1;
	}
	fread(&sig, sizeof(u32), 1, in);
	if(sig == 0x5053507E) // ~PSP - still encrypted
		return !printf("The input file is still encrypted.\nPlease pass it through PRX Decrypter to decrypt it for editing.\n")-1;
	else if(sig == 0x0F803000) // old version
		return !printf("This pspbtcnf.bin is from an older firmware.\n")-1;
	else if(sig == 0x0F803001) // current version
		return 0;
	fclose(in);
	return -1;
}

int split_btcnf(char *oldfile)
{
	FILE *in, *out;
	char filename[256];
	BtcnfHeader head;
	ModeEntry *modes;
	ModuleEntry module;
	int x,y;

	if(check_sig(oldfile)) return -1;
	in = fopen(oldfile, "r");
	if(!in)
	{
		return !printf("Input file %s not found!\n", oldfile)-1;
	}

	fread(&head, sizeof(BtcnfHeader), 1, in);
	modes = (ModeEntry *)malloc(head.nmodes*sizeof(ModeEntry));
	fseek(in, head.modestart, SEEK_SET);

	printf("Loading module list... ");
	for(x = 0; x < head.nmodes; x++)
		fread(&modes[x], sizeof(ModeEntry), 1, in);
	printf("done.\nWriting module list... ");
	out = fopen(BTCNF, "w");
	for(x=0;x<head.nmodules;x++)
	{
		fseek(in, head.modulestart+0x20*x, SEEK_SET);
		fread(&module, sizeof(ModuleEntry), 1, in);
		fseek(in, module.stroffset+head.modnamestart, SEEK_SET);
		fgets(filename, 255, in);
		if(module.loadmode & TYPE_DOLLAR) fputs("$", out);
		if(module.loadmode & TYPE_PERCENT) fputs("%", out);
		if(module.loadmode & TYPE_TWOPERCENT) fputs("%%", out);
		fprintf(out, "%s ", filename);

		if(module.flags & SECT_VSH)
			fprintf(out, "V");
		if(module.flags & SECT_GAME)
			fprintf(out, "G");
		if(module.flags & SECT_UPDATER)
			fprintf(out, "U");
		if(module.flags & SECT_POPS)
			fprintf(out, "P");
		if(module.flags & SECT_LICENSE)
			fprintf(out, "L");
		if(module.flags & SECT_APP)
			fprintf(out, "A");
		if(module.flags & SECT_UMDEMU)
			fprintf(out, "E");
		if(module.flags & SECT_MLNAPP)
			fprintf(out, "M");

		fprintf(out, "\n");
	}
	printf("done.\nOutput saved as 'pspbtcnf.txt'.\n");
	fclose(out);
	fclose(in);
	free(modes);
	return 0;
}

int build_btcnf(char *oldfile)
{
	FILE *in, *out;
	BtcnfHeader head;
	ModeEntry mode_entry;
	ModuleEntry **modules;
	int dump;
	char filename[256];
	char **prxfiles;
		int loadmode, flags;
		int j;
		char cflags[7];
		char c;
	int i;
	int modules_count=0, max_modules=0;
		char sflags[256];
	int module_list_length = 0;
	int len;
	char *outf;
		int found=0;

	in = fopen(BTCNF, "r");
	if(!in)
	{
		return !printf("Input file " BTCNF " not found!\n")-1;
	}

	

	outf = (char *)malloc(strlen(oldfile)+5);
	sprintf(outf, "%s.new", oldfile);
	fcopy(oldfile, outf);

	in = fopen(BTCNF, "r");

	printf("Counting modules... ");
	while(!feof(in))
	{
		memset(filename, 0, 256);

		fscanf(in, "%s %05X", filename, &dump);
		if(strlen(filename) == 0) break;
		max_modules++;
	}

	printf("done.\nLoading module list...");
	modules = (ModuleEntry **)malloc(max_modules * sizeof(ModuleEntry *));

	len = ftell(in);
	fseek(in, 0, SEEK_SET);

	prxfiles = (char **)malloc(max_modules * sizeof(char *));
	while(!feof(in))
	{
		cflags[6]='\0';
		memset(filename, 0, 256);

		
		fscanf(in, "%s %s\n", filename, sflags);

		if(strlen(filename) == 0) break;
		flags = 0;
		for(j = 0; j < strlen(sflags); j++)
		{
			switch(sflags[j])
			{
				case 'v':
				case 'V':
					flags |= SECT_VSH;
					continue;
				case 'g':
				case 'G':
					flags |= SECT_GAME;
					continue;
				case 'u':
				case 'U':
					flags |= SECT_UPDATER;
					continue;
				case 'p':
				case 'P':
					flags |= SECT_POPS;
					continue;
				case 'l':
				case 'L':
					flags |= SECT_LICENSE;
					continue;
				case 'a':
				case 'A':				
					flags |= SECT_APP;
					continue;
				case 'e':
				case 'E':				
					flags |= SECT_UMDEMU;
					continue;
				case 'm':
				case 'M':				
					flags |= SECT_MLNAPP;
					continue;
				default:
					continue;
			}
		}

		max_modules++;

		if(!strncmp(filename, "$%%", 3))
		{
			strncpy(filename, filename+3, strlen(filename)-2);
			loadmode = 0x8004;
		}
		else if(!strncmp(filename, "$%", 2))
		{
			loadmode = 0x8002;
			strncpy(filename, filename+2, strlen(filename)-1);
		}
		else if(!strncmp(filename, "$", 1))
		{
			loadmode = 0x8001;
			strncpy(filename, filename+1, strlen(filename));
		}
		else if(!strncmp(filename, "%%", 2))
		{
			loadmode = 4;
			strncpy(filename, filename+2, strlen(filename)-1);
		}
		else if(!strncmp(filename, "%", 2))
		{
			loadmode = 2;
			strncpy(filename, filename+1, strlen(filename));
		}
		else loadmode = 1;

		for(i=0;i < modules_count; i++)
			if(!strcmp(prxfiles[i], filename))
			{
				found = 1;
				if(modules[i]->flags != flags) found = 0;
			}
		if(!found)
		{
			modules[modules_count] = (ModuleEntry *)malloc(sizeof(ModuleEntry));
			modules[modules_count]->stroffset = module_list_length;
			modules[modules_count]->loadmode = loadmode;
			modules[modules_count]->reserved = 0;
			modules[modules_count]->reserved2 = 0;
			modules[modules_count]->flags=flags;
			memset(modules[modules_count]->hash, 0, 16);

			prxfiles[modules_count] = (char *)malloc(strlen(filename)+1);
			memcpy(prxfiles[modules_count], filename, strlen(filename)+1);
			module_list_length+=strlen(filename)+1;
			modules_count++;
		}
	}

	printf("done.\nWriting output file... ");


	out = fopen(outf, "w");
	head.signature = 0x0F803001;
	head.devkit = 0x03090010;
	head.unknown[0]=0x6B8B4567;
	head.unknown[1]=0x327B23C6;
	head.modestart=0x40;
	head.nmodes=7;
	head.unknown2[0]=0x643C9869;
	head.unknown2[1]=0x66334873;
	head.modulestart=0x120;
	head.nmodules=modules_count;
	head.unknown3[0]=0x74B0DC51;
	head.unknown3[1]=0x19495CFF;
	head.modnamestart=head.modulestart+0x20*head.nmodules;
	head.modnameend=head.modnamestart+module_list_length;
	head.unknown4[0]=0x2AE8944A;
	head.unknown4[1]=0x625558EC;
	fwrite(&head, sizeof(head), 1, out);

	mode_entry.maxsearch=modules_count;
	mode_entry.searchstart=0;
	mode_entry.modeflag = SECT_VSH;
	mode_entry.mode2 = 2;
	memset(mode_entry.reserved, 0, 5*sizeof(int));
	fwrite(&mode_entry, sizeof(ModeEntry), 1, out);

	mode_entry.maxsearch=modules_count;
	mode_entry.searchstart=0;
	mode_entry.modeflag = SECT_GAME;
	mode_entry.mode2 = 1;
	memset(mode_entry.reserved, 0, 5*sizeof(int));
	fwrite(&mode_entry, sizeof(ModeEntry), 1, out);

	mode_entry.maxsearch=modules_count;
	mode_entry.searchstart=0;
	mode_entry.modeflag = SECT_UPDATER;
	mode_entry.mode2 = 3;
	memset(mode_entry.reserved, 0, 5*sizeof(int));
	fwrite(&mode_entry, sizeof(ModeEntry), 1, out);

	mode_entry.maxsearch=modules_count;
	mode_entry.searchstart=0;
	mode_entry.modeflag = SECT_POPS;
	mode_entry.mode2 = 4;
	memset(mode_entry.reserved, 0, 5*sizeof(int));
	fwrite(&mode_entry, sizeof(ModeEntry), 1, out);

	mode_entry.maxsearch=modules_count;
	mode_entry.searchstart=0;
	mode_entry.modeflag = SECT_APP;
	mode_entry.mode2 = 6;
	memset(mode_entry.reserved, 0, 5*sizeof(int));
	fwrite(&mode_entry, sizeof(ModeEntry), 1, out);

	mode_entry.maxsearch=modules_count;
	mode_entry.searchstart=0;
	mode_entry.modeflag = SECT_UMDEMU;
	mode_entry.mode2 = 7;
	memset(mode_entry.reserved, 0, 5*sizeof(int));
	fwrite(&mode_entry, sizeof(ModeEntry), 1, out);

	mode_entry.maxsearch=modules_count;
	mode_entry.searchstart=0;
	mode_entry.modeflag = SECT_MLNAPP;
	mode_entry.mode2 = 8;
	memset(mode_entry.reserved, 0, 5*sizeof(int));
	fwrite(&mode_entry, sizeof(ModeEntry), 1, out);

	for(i=0; i<modules_count; i++)
	{
		if(modules[i] != NULL)
		{
			if(modules[i]->loadmode & 1)
			{
				fwrite(modules[i], sizeof(ModuleEntry), 1, out);
				free(modules[i]);
			}
		}
	}

	for(i=0; i<modules_count; i++)
	{
		if(modules[i] != NULL)
		{
			if(modules[i]->loadmode & (2 | 4) )
			{
				fwrite(modules[i], sizeof(ModuleEntry), 1, out);
				free(modules[i]);
			}
		}
	}
/*
	for(i=0; i<modules_count; i++)
	{
		if(modules[i] != NULL)
		{
			if(modules[i]->loadmode & 4)
			{
				fwrite(modules[i], sizeof(ModuleEntry), 1, out);
				free(modules[i]);
			}
		}
	}
*/
	for(i=0; i<modules_count; i++)
	{
		fputs(prxfiles[i], out);
		fwrite("\0"+1, 1, 1, out);
		free(prxfiles[i]);
	}
	fclose(out);
	free(prxfiles);
	free(modules);
	printf("done.\nNew pspbtcnf file saved as '%s'.\n", outf);
	free(outf);
	return 0;	
}
